#include <assert.h>
#include <check.h>
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "include/ds485-socket-server.h"
#include "digitalSTROM/dsuid/dsuid.h"

#include "tests/lib_simple_client.h"

static struct dS485_server *server;
static pthread_t tid;
static int port = 8445;
static int server_ready = 0;

static ds485_container_t last_pkt;

static pthread_mutex_t recv_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t recv_cond = PTHREAD_COND_INITIALIZER;
static pthread_cond_t ready_cond = PTHREAD_COND_INITIALIZER;

static dsuid_t dsuid1 = { { 0x35, 0x04, 0x17, 0x5f, 0xe0, 0x01, 0x00, 0x00,
    0x00, 0x00, 0x02, 0x03, 0x0a, 0x0b, 0x0c, 0x0d, 0x00 } };
static dsuid_t dsuid2 = { { 0x35, 0x04, 0x17, 0x5f, 0xe0, 0x01, 0x00, 0x00,
    0x00, 0x01, 0x02, 0x03, 0x0a, 0x0b, 0x0c, 0x0d, 0x00 } };

static int receive_data(struct dS485_server *server,
                        const ds485_container_t* packet)
{
    (void)server;
    (void)packet;

    printf("%s\n", __func__);
    pthread_mutex_lock(&recv_mutex);
    memcpy(&last_pkt, packet, sizeof(*packet));
    pthread_cond_signal(&recv_cond);
    pthread_mutex_unlock(&recv_mutex);
    return 0;
}

static int receive_long_data(struct dS485_server *server,
                             const ds485n_packet_t* packet)
{
    (void)server;
    (void)packet;

    printf("%s\n", __func__);
    pthread_mutex_lock(&recv_mutex);
    pthread_cond_signal(&recv_cond);
    pthread_mutex_unlock(&recv_mutex);
    return 0;
}

static int server_status(struct dS485_server *server, int val)
{
    (void)server;
    (void)val;
    printf("%s\n", __func__);
    pthread_mutex_lock(&recv_mutex);
    server_ready = 1;
    pthread_cond_signal(&ready_cond);
    pthread_mutex_unlock(&recv_mutex);
    return 0;
}

void *server_thread(void *arg)
{
    struct dS485_server *server = arg;

    printf("server_thread start\n");
    (void)ds485_server_listen(server, port, false);

    return NULL;
}

struct dS485_server_callbacks cb = {
    .receive_data = receive_data,
    .receive_long_data = receive_long_data,
    .status_change = server_status,
};

static void start_server()
{
    int ret;

    server = ds485_server_init(&cb, true);
    assert(server);

    ret = pthread_create(&tid, NULL, server_thread, server);
    if (ret < 0) {
        ck_abort_msg("failed to create server");
    }

    pthread_mutex_lock(&recv_mutex);
    if (!server_ready) {
        pthread_cond_wait(&ready_cond, &recv_mutex);
    }
    pthread_mutex_unlock(&recv_mutex);
}

static void teardown_server()
{
    (void)ds485_server_abort(server);
}

START_TEST (test_send_client)
{
    ds485_container_t c;
    int s, ret, bytes_available;

    s = connect_server(port);
    ck_assert_int_ge(s, 0);

    memset(&c, 0, sizeof(c));
    memcpy(&c.sourceId, &dsuid1, sizeof(dsuid1));
    memcpy(&c.destinationId, &dsuid2, sizeof(dsuid2));

    c.length = 1;
    c.data[0] = 1;

    ret = send_container(s, &c);
    ck_assert_int_eq(ret, 0);

    pthread_mutex_lock(&recv_mutex);
    pthread_cond_wait(&recv_cond, &recv_mutex);
    pthread_mutex_unlock(&recv_mutex);

    ck_assert(dsuid_equal(&dsuid1, &last_pkt.sourceId));
    ck_assert(dsuid_equal(&dsuid2, &last_pkt.destinationId));

    // now send the reply
    memset(&c, 0, sizeof(c));
    memcpy(&c.sourceId, &dsuid2, sizeof(dsuid_t));
    memcpy(&c.destinationId, &dsuid1, sizeof(dsuid_t));
    c.transactionId = last_pkt.transactionId;
    /* pkt.length == 0 => DS485_CONTAINER_SIZE + 3(TLV) */

    ds485_server_reply(server, &c);
    ioctl(s, FIONREAD, &bytes_available);
    ck_assert_int_eq(bytes_available, DS485_CONTAINER_SIZE + 3);
    (void)recv_container(s, &c);

    ck_assert(dsuid_equal(&dsuid1, &c.destinationId));
    ck_assert(dsuid_equal(&dsuid2, &c.sourceId));

    close(s);
}
END_TEST

Suite *f_suite()
{
    Suite *s;
    TCase *tc_v1_source_id;

    s = suite_create("ds485n");
    tc_v1_source_id = tcase_create("v1 packet with valid source id");

    tcase_add_checked_fixture(tc_v1_source_id, start_server, teardown_server);

    tcase_add_test(tc_v1_source_id, test_send_client);
    suite_add_tcase(s, tc_v1_source_id);

    return s;
}

int main()
{
    Suite *s;
    SRunner *sr;
    int failed;

    s = f_suite();
    sr = srunner_create(s);

    srunner_run_all(sr, CK_VERBOSE);
    failed = srunner_ntests_failed(sr);
    srunner_free(sr);

    return (failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
